% nibbled.m
%
% Show that it is not possible to cover a chess board (8x8) with 31 1x2
% dominoes when one pair of opposite corners is missing.
%
% Grackle output:
%
% rank of LLL lattice: 112
% Starting lattice search (Yices?  OFF)
% number of iterations: 1
% number of prunes:     0
% BLT checkSat completed
% 
% msg =
% 
%     UNSAT
% 
% msg =
% 
%     Time ellapsed:
% 
% time =
% 
%     0.8580
% 
% Execution complete


% Use BLT as the solver backend
mss_config('default_solver', 'blt');
% Uncomment to use Yices as the solver backend
% mss_config('default_solver', 'yices');

% positions on an 8 by 8 rectangular grid
m = 20;  % nrows
n = 20;  % ncols

% positions to cover on the board
board = ones(m, n, 'uint8');
board(1,1) = 0;
board(m,n) = 0;

% uncomment to get a feasible board
board(1,2) = 0;
board(m,n-1) = 0;

ndom = sum(sum(board))/2;  % # dominoes needed to cover

% v(i,j) represents the choice of using a vertical domino
% whose top half covers position (i,j); the h(i,j) similarly
% represent horizontal dominoes whose left half covers (i,j)
v = symbolic('v', m-1, n, 'integer');
h = symbolic('h', m, n-1, 'integer');


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Build a predicate

% 'p' is a predicate we build up and check for satisfiability at the end
p = true;

% {0,1} constraints on variables: enforces the constraint that
% each potential domino is either present or not
for i = 1:(m-1)
    for j = 1:(n)
        p = p && 0 <= v(i,j) && v(i,j) <= 1;
    end
end
for i = 1:(m)
    for j = 1:(n-1)
        p = p && 0 <= h(i,j) && h(i,j) <= 1;
    end
end

% covering constraints for the chess board, we require every board square is
% covered by at least on domino
for i = 1:m
    for j = 1:n
        if board(i,j)
            % q represents the number of times a given position is covered
            q = 0;
            % vertical from above
            if i > 1
                q = q + v(i-1,j);
            end
            % vertical from below
            if i+1 <= m
                q = q + v(i,j);
            end
            % horizontal from the left
            if j > 1
                q = q + h(i,j-1);
            end
            % horizontal from the right
            if j+1 <= n
                q = q + h(i,j);
            end
            p = p && 1 <= q && q <= 4;
        end
    end
end

% Finally, require that we use at most 'ndom' dominoes. This, together with the
% covering condition implies that we must use *exactly* 'ndom' dominoes.
s = sum(sum(v)) + sum(sum(h));
p = p && 0 <= s && s <= ndom;


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Check the predicate for satisfiability

t0 = clock;
[r,c] = check_sat(p);
if r == 1
    msg = 'Found feasible solution'
else
    msg = 'UNSAT'
end
t1 = clock;
msg = 'Time ellapsed:'
time = etime(t1, t0)
