% This script verifies the correctness of the Boyer-Moore
% majority vote algorithm for individual fixed sizes.
% Sizes up to about 20 are quite feasible; after that
% solvers start to take exponentially longer to verify.
%
% Moreover, the solvers are sensitive to the type chosen for the
% data elements in the array.  For example, Yices and Z3 take
% significantly longer to do the verification if the elements are
% bounded integers than if they are integers or reals.  We
% conjecture that this is because this activates the bitvector theory.
%
% For details of the original algorithm, see:
%
%   "MJRTY - A Fast Majority Vote Algorithm." Robert S. Boyer and
%   J. Strother Moore. In _Automated Reasoning: Essays in Honor of
%   Woody Bledsoe_.  Kulwer, Dordrecht, The Netherlands. 1991. Pages 105--117.

mss_config('default_solver', 'yices');
sz = 25;
val_type = 'integer';

% produce a linear array of `sz` variables
x = symbolic('x', 1, sz, val_type );

% suppose that `maj` is the majority element of `x`
maj = symbolic( 'maj', val_type );
assume( sum( x == maj ) > length(x)/2 );

% Now calculate the majority using the MJRTY algorithm
c = find_majority(x);

assert( c == maj, 'find_majority postcondition' );

remaining = prove_obligations;
if remaining == 0
    'Verification comlete!'
else
    'Unproved goals remain!'
end
