{-# LANGUAGE ParallelListComp #-}
{-# LANGUAGE ScopedTypeVariables #-}

module Main where

import Control.Applicative
import Control.Exception
import System.IO
import System.Environment

import Data.Text (Text)
import Lang.Crucible.Solver.SimpleBackend.Yices
import Lang.Crucible.Solver.SimpleBackend.SMTWriter

windowSize :: Int
windowSize = 30

inputLimit :: Rational
inputLimit = 127.0

sumLimit :: Rational
sumLimit = 10.0 * fromIntegral windowSize

initialInput :: [Rational]
initialInput = []

type T = Term (Connection ())
type F = Formula (Connection ())
type S = [T]

inputInvariant :: T -> F
inputInvariant t = rationalTerm (-inputLimit) .<= t .&& t .<= rationalTerm inputLimit

outputInvariant :: T -> F
outputInvariant t = rationalTerm (-sumLimit) .<= t .&& t .<= rationalTerm sumLimit

transition :: S -> T -> S -> F
transition s t s' = assert (length s == length s') $
  outputInvariant (foldr1 (+) (s++[t])) .&&
  foldr1 (.&&) [ a .== b | a <- drop 1 (s++[t])
                         | b <- s'
                         ]

declareState :: Connection () -> IO S
declareState c = do
  s <- sequence [ fromText <$> mkFreeVar c SMT_RealType | _i <- [0 .. windowSize-2] ]
  sequence_ [ assumeFormula c (inputInvariant x) | x <- s ]
  return s

getFreshState :: Connection () -> IO [Text]
getFreshState c = do
  sequence [ getFreshName c | _i <- [0 .. windowSize-2] ]

mkInitialState :: Connection () -> IO S
mkInitialState c = do
   vs <- reverse <$> sequence
         [ v
         | _i <- [0 .. windowSize-2]
         | v  <- (reverse (map (return . rationalTerm) initialInput))
                  ++ repeat (fromText <$> mkFreeVar c SMT_RealType)
         ]
   sequence_ [ assumeFormula c (inputInvariant v) | v <- vs ]
   return vs

buildUnreachable :: Connection () -> Int -> S -> S -> IO ()
buildUnreachable c 0 sinitial sfinal = do
   let neqs = [ notExpr (x .== y)
              | x <- sinitial
              | y <- sfinal
              ]
   assumeFormula c (foldr1 (.||) neqs)

buildUnreachable c k sinitial sfinal = do
   js' <- sequence [ getFreshName c | _i <- [0 .. k-1] ]
   ss' <- sequence [ getFreshState c | _i <- [0 .. k-2] ]

   let e = goalFormula (map fromText js') (sinitial : map (map fromText) ss' ++ [sfinal])
   assertForall c [ (v, yicesType SMT_RealType) | v <- js' ++ concat ss' ] e

goalFormula :: [T] -> [S] -> F
goalFormula js ss =
  foldr1 (.||) $ map notExpr $
      [ inputInvariant j | j <- js ]
      ++
      [ inputInvariant x | s <- ss, x <- s ]
      ++
      [ transition s j s'
      | s <- ss
      | j <- js
      | s' <- drop 1 ss
      ]

main :: IO ()
main = do
  [k_str] <- getArgs
  let (k::Int) = read k_str

  c <- newConnection stdout 
  js <- sequence [ mkFreeVar c SMT_RealType | _i <- [0 .. k] ]

  sinitial <- mkInitialState c
  ss <- sequence [ declareState c | _i <- [0 .. k] ]
  let sfinal = last ss

  sequence_ [ assumeFormula c (inputInvariant j)  | j <- map fromText js ]

  sequence_ [ assumeFormula c (transition s j s')
            | j  <- map fromText js
            | s  <- sinitial : ss
            | s' <- ss
            ]

  sequence_ [ buildUnreachable c i sinitial sfinal
            | i <- [0 .. k]
            ]

  checkExistsForall c
  sequence_ [ eval c j | j <- js ]

