{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE ParallelListComp #-}
{-# LANGUAGE ScopedTypeVariables #-}

module Main where

import qualified Data.Foldable as Fold
--import Control.Applicative
import Control.Exception
import Control.Concurrent
import Control.Monad
import System.IO
import System.Exit
--import System.Environment
import Data.Sequence (Seq)
import qualified Data.Sequence as Seq

import qualified System.IO.Streams as Streams
import qualified System.IO.Streams.Attoparsec as Streams

--import Data.Text (Text)
import Data.ByteString (ByteString)
import Data.Ratio
import Lang.Crucible.Solver.SimpleBackend.Yices
import Lang.Crucible.Solver.SimpleBackend.SMTWriter
import Lang.Crucible.Solver.ProcessUtils
import Lang.Crucible.Utils.SExp
import Lang.Crucible.Utils.Streams

shortWindowSize :: Int
shortWindowSize = 3

windowSize :: Int
windowSize = 30

inputLimit :: Int
inputLimit = 127

sumLimit :: Int
sumLimit = 10 * windowSize

k_limit :: Int
--k_limit = 54
k_limit = 49

type T = Term (Connection ())
type F = Formula (Connection ())
type S = [T]

inputInvariant :: T -> F
inputInvariant t = rationalTerm (fromIntegral (-inputLimit)) .<= t .&&
               t .<= rationalTerm (fromIntegral inputLimit)

outputInvariant :: T -> F
outputInvariant t = rationalTerm (fromIntegral (-sumLimit)) .<= t .&&
                t .<= rationalTerm (fromIntegral sumLimit)

slideWindow :: Num a => Int -> [a] -> [a]
slideWindow w = \xs -> go (Seq.fromList $ take w xs) (drop w xs)
 where go :: Num a => Seq a -> [a] -> [a]
       go buf []     = [ Fold.foldl1 (+) buf ]
       go buf (x:xs) =
            let buf' = (Seq.drop 1 buf) Seq.|> x
             in (Fold.foldl1 (+) buf) : (go buf' xs)

logYices :: String -> IO ()
logYices _ = return ()
--logYices str = putStrLn str

packOneString :: String -> IO ()
packOneString s = do
  withProcessHandles "yices" [] Nothing $ \(pin, pout, perr, ph) -> do
    c <- newConnection pin
    outstr <- Streams.handleToInputStream pout
    -- Log stderr to output.
    err_stream <- Streams.handleToInputStream perr
    _ <- forkIO $ logErrorStream err_stream logYices

    let n = 2*k_limit + length s
    vs <- sequence [ mkFreeVar c SMT_IntegerType | _ <- [1..n] ]
    let xs = map fromText vs
    sequence_ [ assumeFormula c (inputInvariant x) | x <- xs ]
    let ys = slideWindow shortWindowSize xs
    let zs = slideWindow windowSize xs
    sequence_ [ assumeFormula c (y .== rationalTerm (fromIntegral (fromEnum char)))
              | char <- s
              | y <- drop k_limit ys
              ]
    sequence_ [ assumeFormula c (outputInvariant z)
              | z <- zs
              ]
    check c
    sat_result <- try $ Streams.parseFromStream parseNextWord outstr
    --putStr $ s++" : "
    case sat_result of
      Left (Streams.ParseException msg) -> do
        fail $ "Could not parse result from yices: " ++ msg
      Right "unsat" -> putStrLn "unsat"
      Right "sat" -> return ()
          --putStrLn "sat"
          --sequence_ [ print =<< yicesRealValue =<< requestValue c outstr v | v <- vs ]
      Right msg -> do
         putStrLn "unknown"
         putStrLn msg
    _ <- waitForProcess ph
    return ()

appendString :: Connection s -> Streams.InputStream ByteString -> Seq Int -> String -> IO (Seq Int)
appendString c outstr q str = inFreshFrame c $ do
    let n = k_limit + length str + shortWindowSize
    let prefix = Fold.toList $ Seq.reverse $ Seq.take windowSize $ Seq.reverse $ q
    vs <- sequence [ mkFreeVar c SMT_IntegerType | _ <- [1..n] ]
    sequence_ [ assumeFormula c (inputInvariant (fromText v)) | v <- vs ]

    let xs = map (rationalTerm . fromIntegral) prefix ++ map fromText vs
    let ys = slideWindow shortWindowSize xs
    let zs = slideWindow windowSize xs
    sequence_ [ assumeFormula c (y .== rationalTerm (fromIntegral (fromEnum char)))
              | char <- str
              | y <- drop (length prefix + k_limit) ys
              ]
    sequence_ [ assumeFormula c (outputInvariant z)
              | z <- zs
              ]
    check c
    sat_result <- try $ Streams.parseFromStream parseNextWord outstr
    case sat_result of
      Left (Streams.ParseException msg) -> do
          fail $ "Could not parse result from yices: " ++ msg
      Right "sat" -> do
          --putStrLn $ unwords ["sat:", str]
          rs <- sequence [ yicesRealValue =<< requestValue c outstr v | v <- vs ]
          let is = [ assert (denominator r == 1) (fromInteger $ numerator r) | r <- rs ]
          return (q Seq.>< Seq.fromList is)
      Right "unsat" -> do
          die $ unwords ["unsat! while packing: ", str]
      Right msg -> do
          die $ unwords ["unknown! while packing: ", str, msg]

main :: IO ()
main = do
  hSetBuffering stdout LineBuffering
  ws <- lines <$> (hGetContents =<< openFile "shuffle.txt" ReadMode)

  is <- withProcessHandles "yices" [] Nothing $ \(pin, pout, perr, ph) -> do
    c <- newConnection pin
    outstr <- Streams.handleToInputStream pout
    -- Log stderr to output.
    err_stream <- Streams.handleToInputStream perr
    _ <- forkIO $ logErrorStream err_stream logYices

    is <- foldM (appendString c outstr) Seq.empty ws
    _ <- waitForProcess ph
    return is

  let xs = Fold.toList is
  let ys = slideWindow shortWindowSize xs
  let zs = slideWindow windowSize xs
  print xs
  print ys
  print zs
  verifyInputBound xs
  verifyStrings ws ys
  verifyOutputBound zs
  print $ unwords ["length xs:", show $ length xs]

verifyInputBound :: [Int] -> IO ()
verifyInputBound xs =
   if Fold.all ok xs then return () else fail "Failed input bound check"
 where ok x = -inputLimit <= x && x <= inputLimit

verifyStrings :: [String] -> [Int] -> IO ()
verifyStrings [] _ = return ()
verifyStrings (w:ws) ys
  | null ys = fail $ unwords ["Failed embedded strings check!", show (w:ws)]
  | and [ fromEnum c == y
        | c <- w
        | y <- ys
        ]
     = verifyStrings ws (drop (length w) ys)
  | otherwise = verifyStrings (w:ws) (drop 1 ys)

verifyOutputBound :: [Int] -> IO ()
verifyOutputBound zs = do
   if Fold.all ok zs then return () else fail "Failed output bound check"
 where ok x = -sumLimit <= x && x <= sumLimit
