% This is the CAMUS-based search procedure for the 2D
% string-packing problem.

% Parameters to the problem
len        =   100; % how long are the sides of the input
width      =     3; % how wide is the narrow sliding avg. box
extra      =     5; % how much wider is the wide sliding avg. box
maxsearch  =   250; % how many iterations to search before resetting
limit      = 0.001; % what is the limit for the wide sliding average?
searchinc  =     1; % how many spaces to skip when we search 
inputlimit =   250; % maximum distance from 0 an input value can be

% the strings we want to pack
segments  = { 'hello!asdf' ...
            , 'whatup'  ...
            , 'yall' ...
            , 'howdy', 'pardner' ...
            , 'WHATDO', 'YOUWANT' ...
            , 'howareyou' ...
            , 'my man' ...
            };
numsegs   = size(segments, 2);

% The symbolic input vector we are searching for
input     = symbolic('input', len, len, 'integer');

% The vector of locations at which we place the strings
sym_gaps   = symbolic('gaps', 2, numsegs, 'integer');

% symbolically calculate the narrow sliding window avg.
y = slide2d( input, width, 1 );
y_len = size(y, 2);

% symbolically calculate the wide sliding window avg.
z = slide2d( input, width+extra, 1 );

%% Now begin the location search algorithm

%  1) The outer loop gradually increases the 'minsep' value which
%     determines the minimum amount of separation between packed
%     strings.  It continues until eventually we find a solution.
minsep = 0;
c = false;
more_models = true;
while ( not(c) & minsep < len )
    minsep = minsep + searchinc;
    sym_locs = locs_from_gaps2d(segments, sym_gaps);

    locs_ok = true;
    for i = 1:numsegs
        % every location is in range
        locs_ok = locs_ok & minsep <= sym_locs(1,i) ...
                          & minsep <= sym_locs(2,i) ...
                          & sym_locs(1, i) + minsep <= y_len ...
                          & sym_locs(2, i) + size(segments{i},2) + minsep <= y_len;

        % every string packing location is disjoint from every other
        for j = i+1:numsegs
            ij_disjoint ...
                = sym_locs(1,j) + minsep <= sym_locs(1,i) ...
                | sym_locs(1,i) + minsep <= sym_locs(1,j) ...
                | sym_locs(2,j) + size(segments{j},2) + minsep <= sym_locs(2,i) ...
                | sym_locs(2,i) + size(segments{i},2) + minsep <= sym_locs(2,j);
            locs_ok = locs_ok & ij_disjoint;
        end
    end

    locmodels = 0;
    [more_models loc_model cont] = get_solutions( locs_ok );
    
    while( more_models & locmodels < maxsearch )
        locmodels = locmodels + 1;
        gaps = loc_model.gaps;
        locs = locs_from_gaps2d( segments, loc_model.gaps );

        % print the current state of the search
        minsep, locmodels, gaps, locs

        % Here we build the assertion that all the strings
        % appear in the output, and that the jump pointers
        % correctly link up to the next string in the sequence.
        clauses = ones(1, numsegs);
        for j = 1:numsegs
            segsize = size(segments{j}, 2);
            % every character in the string appears in order at the
            % chosen location
            for k = 1:segsize
                clauses(j) = clauses(j) & (y(locs(1,j), locs(2,j) + k - 1) == segments{j}(k));
            end
            % The next two values give the "jump" values that
            % indicate where to find the next string.  The final
            % string has jump values of 0.
            if( j < numsegs )
                clauses(j) = clauses(j) & y( locs(1,j), locs(2,j) + segsize ) == (locs(1,j+1) - locs(1,j));
                clauses(j) = clauses(j) & y( locs(1,j), locs(2,j) + segsize + 1) == (locs(2,j+1) - locs(2,j));
            else
                clauses(j) = clauses(j) & (y( locs(1,j), locs(2,j) + segsize ) == 0 );
                clauses(j) = clauses(j) & (y( locs(1,j), locs(2,j) + segsize + 1 ) == 0 );
            end
        end

        % Here we bound the size of the larger sliding average:
        % this has the general effect of forcing the input
        % vector to be roughly centered around 0.
        p = all(all( z <= limit )) & all(all( z >= -limit ));
        
        % We also directly bound the magnitude of the input values
        p = p & all(all( input <= inputlimit )) & all(all( input >= -inputlimit ));

        % Try to find a solution
        [c r] = check_sat( p & all( clauses ) );
        c

        if( c )
            break;
        end

        % If the current location vector is not a solution,
        % calculate some unsat cores.
        %cores = find_mus( p, clauses )
        %cores = find_all_mus( p, clauses, 10 )
        cores = find_all_mus( p, clauses )

        % optional, verify that the calculated cores are actually
        %   unsatisfiable and minimal
        % for i = 1:size(cores,1)
        %    verify_core( p, clauses, cores(i,:) );
        % end

        % Using the unsat cores, compute a constraint that
        % hopefully will push us toward a solution.  Examine every
        % core generated above and increase the minimum distance between
        % at least one pair of strings occuring in each core.
        contpred = true;
        haspairs = false;
        for coreidx = 1:size(cores,1)
            core = cores(coreidx,:);
            
            haspairs = false;
            corepred = false;
            for i = 1:numsegs
                if core(i)
                    for j = i+1:numsegs
                        if core(j)
                            haspairs = true;
                            if ( abs(locs(1,j) - locs(1,i)) < abs(locs(2,j) - locs(2,i)) )
                                corepred = corepred | ...
                                    abs(sym_locs(1,j) - sym_locs(1,i)) >= ...
                                    abs(locs(1,j) - locs(1,i)) + searchinc;
                            else
                                corepred = corepred | ...
                                    abs(sym_locs(2,j) - sym_locs(2,i)) >= ...
                                    abs(locs(2,j) - locs(2,i)) + searchinc;
                            end
                        end
                    end
                end
            end

            % If no pair exists, push the troublesome string further from its closest boundary
            if not(haspairs)
                for i = 1:numsegs
                    if core(i)
                        sym_min = min( [ sym_locs(1,i), ...
                                         len - sym_locs(1,i), ...
                                         sym_locs(2,i), ...
                                         len - sym_locs(2,i) - size(segments{i},2)] );
                        concrete_min = min( [ locs(1,i), ...
                                            len - locs(1,i), ...
                                            locs(2,i), ...
                                            len - locs(2,i) - size(segments{i},2) ] );
                        corepred = sym_min >= searchinc + concrete_min;
                        break;
                    end
                end
            end

            contpred = contpred & corepred;
        end

        % get another location vector and try again
        [more_models loc_model cont] = cont( contpred );
    end
end

% If we succeed in finding a solution, print it out and calculate
% the intermediate values of interest.
if c 
    r.input
    gaps
    locs
    
    % Print the results of the narrow sliding average
    q = slide2d( r.input, width, 1 )

    % Print the results of the wide sliding average
    w = slide2d( r.input, width+extra, 1 )

    % Print the segments of the output where we expect to find
    % the desired strings packed
    for i = 1:numsegs
        q( locs(1,i),  (locs(2,i) : (locs(2,i)+size(segments{i},2)+1)) )
    end
else
    'No solution found!!'
end
