% This is a modification of the incremental greedy search algorithm for
% the 2D string-packing problem.  

% Parameters to the problem
len        =   100; % how long are the sides of the input
width      =     3; % how wide is the narrow sliding avg. box
extra      =     5; % how much wider is the wide sliding avg. box
maxsearch  =   250; % how many iterations to search before resetting
limit      = 0.001; % what is the limit for the wide sliding average?
searchinc  =     1; % how many spaces to skip when we search 
inputlimit =   250; % maximum distance from 0 an input value can be

% the strings we want to pack
segments  = { 'hello!asdf' ...
            , 'whatup'  ...
            , 'yall' ...
            , 'howdy', 'pardner' ...
            , 'WHATDO', 'YOUWANT' ...
            , 'howareyou' ...
            , 'my man' ...
            };
numsegs   = size(segments, 2);

% The symbolic input vector we are searching for
input     = symbolic('input', len, len, 'integer');

% The vector of locations at which we place the strings
sym_gaps   = symbolic('gaps', 2, numsegs, 'integer');

% symbolically calculate the narrow sliding window avg.
y = slide2d( input, width, 1 );
y_len = size(y, 2);

% symbolically calculate the wide sliding window avg.
z = slide2d( input, width+extra, 1 );

%% Now begin the location search algorithm

%  1) The outer loop gradually increases the 'minsep' value which
%     determines the minimum amount of separation between packed
%     strings.  It continues until eventually we find a solution.
minsep = 0;
c = false;
more_models = true;
while ( not(c) & minsep < len )
    minsep = minsep + searchinc;
    sym_locs = locs_from_gaps2d(segments, sym_gaps);

    locs_ok = true;
    for i = 1:numsegs
        % every location is in range
        locs_ok = locs_ok & minsep <= sym_locs(1,i) ...
                          & minsep <= sym_locs(2,i) ...
                          & sym_locs(1, i) + minsep <= y_len ...
                          & sym_locs(2, i) + size(segments{i},2) + minsep <= y_len;

        % every string packing location is disjoint from every other
        for j = i+1:numsegs
            ij_disjoint ...
                = sym_locs(1,j) + minsep <= sym_locs(1,i) ...
                | sym_locs(1,i) + minsep <= sym_locs(1,j) ...
                | sym_locs(2,j) + size(segments{j},2) + minsep <= sym_locs(2,i) ...
                | sym_locs(2,i) + size(segments{i},2) + minsep <= sym_locs(2,j);
            locs_ok = locs_ok & ij_disjoint;
        end
    end

    [more_models loc_model cont] = get_solutions( locs_ok );

    % Try to pack an increasing number of strings in each iteration
    for i=1:numsegs
        locmodels = 0;

        % Search until we cannot find any satisfying location vector,
        % or until we hit our search limit
        while( more_models & locmodels < maxsearch )
            locmodels = locmodels + 1;
            gaps = loc_model.gaps;
            locs = locs_from_gaps2d( segments, loc_model.gaps );

            % print the current state of the search
            minsep, i, locmodels, gaps, locs

            % Here we build the assertion that all the strings
            % up to string i appear in the output, and that the
            % correctly link up to the next string in the sequence.
            clauses = ones(1, numsegs);
            for j = 1:i
                segsize = size(segments{j}, 2);
                % every character in the string appears in order at the
                % chosen location
                for k = 1:segsize
                    clauses(j) = clauses(j) & (y(locs(1,j), locs(2,j) + k - 1) == segments{j}(k));
                end
                % The next two values give the "jump" values that
                % indicate where to find the next string.  The final
                % string has jump values of 0.
                if( j < numsegs )
                    clauses(j) = clauses(j) & y( locs(1,j), locs(2,j) + segsize ) == (locs(1,j+1) - locs(1,j));
                    clauses(j) = clauses(j) & y( locs(1,j), locs(2,j) + segsize + 1) == (locs(2,j+1) - locs(2,j));
                else
                    clauses(j) = clauses(j) & (y( locs(1,j), locs(2,j) + segsize ) == 0 );
                    clauses(j) = clauses(j) & (y( locs(1,j), locs(2,j) + segsize + 1 ) == 0 );
                end
            end
            
            % Here we bound the size of the larger sliding average:
            % this has the general effect of forcing the input
            % vector to be roughly centered around 0.
            p = all(all( z <= limit )) & all(all( z >= -limit ));
        
            % We also directly bound the magnitude of the input values
            p = p & all(all( input <= inputlimit )) & all(all( input >= -inputlimit ));
            
            % Try to find a solution
            [c r] = check_sat( p & all( clauses ) );
            c
            
            % Break out if we find a solution.  If not,
            % increase the minimum distance for the string we are currently adding
            % from it's nearest string or from the boundary.
            if( c )
                break;
            else
                p_addl = min_dist_2d( sym_locs, len, j ) >= min_dist_2d( locs, len, j ) + searchinc;
                [more_models loc_model cont] = cont( p_addl );
            end
        end
        
        if( not(c) )
            break;
        end
    end
end

if c 
    r.input
    gaps
    locs
    
    % Print the results of the narrow sliding average
    q = slide2d( r.input, width, 1 )

    % Print the results of the wide sliding average
    w = slide2d( r.input, width+extra, 1 )

    % Print the segments of the output where we expect to find
    % the desired strings packed
    for i = 1:numsegs
        q( locs(1,i),  (locs(2,i) : (locs(2,i)+size(segments{i},2)+1)) )
    end
else
    'No solution found!!'
end
